/**
 * The MySensors Arduino library handles the wireless radio link and protocol
 * between your home built sensors/actuators and HA controller of choice.
 * The sensors forms a self healing radio network with optional repeaters. Each
 * repeater and gateway builds a routing tables in EEPROM which keeps track of the
 * network topology allowing messages to be routed to nodes.
 *
 * Created by Henrik Ekblad <henrik.ekblad@mysensors.org>
 * Copyright (C) 2013-2015 Sensnology AB
 * Full contributor list: https://github.com/mysensors/Arduino/graphs/contributors
 *
 * Documentation: http://www.mysensors.org
 * Support Forum: http://forum.mysensors.org
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 *******************************
 *
 * DESCRIPTION
 * This sketch pretends to control a window blind with an Arduino, and
 * uses this example [how to implement a humidity/temperature
 * sensor using DHT11/DHT-22 
 * http://www.mysensors.org/build/humidity] to add more funcionality.
 * Each node is controlling a blind, and has a DHT22 sensor and, maybe,
 * a light sensor. Every 30 seconds the node send to the GW the Tª, H 
 * blind position. 
 * By the time, it's always listening if there's some remote order to
 * move the blind, or if the blind is moved using the wall-mounted 
 * buttons. 
 * In the future more funcionality could be added: a LCD display to 
 * show Tª & H, PIR sensor to control room lightning, an IR sensor in 
 * order to work locally with a remote control...
 */
 
// Enable debug prints
#define SERIAL_DEBUG true
#include "SerialDebug.h"

#define EEPROM_LOCAL_ADDRESS 2 // First free address for sketch static configuration
#define MY_PARENT_NODE_ID 0
#define MY_NODE_ID 2
// Enabled repeater feature for this node
// #define MY_REPEATER_FEATURE

// Enable and select radio type attached
#define MY_RADIO_NRF24
//#define MY_RADIO_RFM69

#include <SPI.h>
#include <MySensor.h>  
#include <DHT.h>  
#include <Bounce2.h>

#define CHILD_ID_HUM 20
#define CHILD_ID_TEMP 21
#define HUMIDITY_SENSOR_DIGITAL_PIN 2
unsigned long SLEEP_TIME = 30000; // Sleep time between reads (in milliseconds)

DHT dht;
float lastTemp;
float lastHum;
MyMessage msgHum(CHILD_ID_HUM, V_HUM);
MyMessage msgTemp(CHILD_ID_TEMP, V_TEMP);

// Parametros para el sensor de luz
int  switch_state;  // Holds the last digital value 
bool existsSensor = false; // Only if there is a light sensor on this node

#define CHILD_ID_L 22   // Id of the sensor child for Light Sensor
// Digital Data: A LOW signal when the light is brighter than the set point.
#define LIGHT_SENSOR_DIGITAL_PIN 4 
//#define LED_INT_PIN    13  // The onboard LED
 
MyMessage msgLuz(CHILD_ID_L,V_STATUS); 

#define RELAY_ON_PIN  7  // Arduino Digital I/O pin number for relay 
#define RELAY_MV_PIN  8  // Arduino Digital I/O pin number for relay 
#define BUTTON_UP_PIN 5  // Arduino Digital I/O pin number for button
#define BUTTON_DOWN_PIN 3 // Arduino Digital I/O pin number for button 

#define CHILD_ID_REL_ON 23   // Id of the sensor child for relay on/off blinds
#define CHILD_ID_REL_MV 24   // Id of the sensor child for relay up/down
#define RELAY_OFF HIGH
#define RELAY_ON LOW

Bounce debouncer_up = Bounce(); // Debouncer relay on/off
int oldValueUP=0;
Bounce debouncer_down = Bounce(); // Debouncer relay up/down
int oldValueDW=0;

// Blinds parameters
MyMessage msg(CHILD_ID_REL_MV,V_LIGHT);
MyMessage msgPos(CHILD_ID_REL_MV,V_DIMMER);
//float positionValue;   // the position value
uint8_t positionValue;   // the position value [0-100]
// Operation times for the blind (could be parametrized!!)
unsigned long OPERATION_TIME = 15500;  

// Read DHT22 values: Temperature & Humidity

void sense() { 
  // Fetch temperatures from DHT sensor
  float temperature = dht.getTemperature();
    if (isnan(temperature)) {
          DEBUG("Failed reading temperature from DHT");
    } else if (temperature != lastTemp) {
        lastTemp = temperature;                  
        send(msgTemp.set(temperature, 1));
        DEBUG("T: ", temperature);
    }
    // Fetch humidity from DHT sensor  
    float humidity = dht.getHumidity();
    if (isnan(humidity)) {
        DEBUG("Failed reading humidity from DHT");
    } else if (humidity != lastHum) {
        lastHum = humidity;
        send(msgHum.set(humidity, 1));
        DEBUG("H: ", humidity);
    }      
}

// Read data from light sensor
void  changeLight(){  
  if (switch_state != digitalRead(LIGHT_SENSOR_DIGITAL_PIN))
  {   // changed !! send to GW
    switch_state = digitalRead(LIGHT_SENSOR_DIGITAL_PIN);       
    send(msgLuz.set(switch_state));   
    DEBUG("Light changed !! new value: ", switch_state);   
  } 
}
void upAction(){  
  unsigned long nowMillis;
  unsigned long startMillis;
  int valueU;
  positionValue = loadState(EEPROM_LOCAL_ADDRESS); 
  boolean changed = debouncer_up.update();
  //uint8_t positionValue;
  if ( changed ) {  // Button pressed
        // save last state
      valueU = debouncer_up.read();
      // compare the buttonState to its previous state
      if (valueU != oldValueUP) {
          // if the current state is HIGH then:
          //if (valueU == HIGH && ((uint8_t)loadState(EEPROM_LOCAL_ADDRESS))<100) {
          if (valueU == HIGH && positionValue<100) {
                  moveUP(); 
                  startMillis = millis();
                  do { 
                      debouncer_up.update();
                      valueU=debouncer_up.read(); 
                      nowMillis = millis();
                  } while ( valueU == HIGH 
                          && ((nowMillis-startMillis)/OPERATION_TIME)<100 
                          && (positionValue+((nowMillis-startMillis)/OPERATION_TIME))<100
                          &&  (nowMillis-startMillis)<OPERATION_TIME ); 
                  saveState(EEPROM_LOCAL_ADDRESS, (uint8_t) (positionValue+(uint8_t)((nowMillis-startMillis)/OPERATION_TIME)));
                  oldValueUP=valueU;
                  moveSTOP();
              } else  {  // LOW to HIGH  (Released UP Button)
                DEBUG("Button UP released after ....", nowMillis-startMillis, "miliseconds" );             
                somethingWrong();              
          
             } 
         } 
  }

}
void downAction(){  
  unsigned long nowMillis;
  unsigned long startMillis;
  int valueD;
  boolean changed = debouncer_down.update();  
  positionValue = loadState(EEPROM_LOCAL_ADDRESS); 
    
  if ( changed ) {  // Button pressed
        // save last state
      valueD = debouncer_down.read();
      // compare the buttonState to its previous state
      if (valueD != oldValueDW) {
          // if the current state is HIGH then:
          //if (valueD == HIGH && loadState(EEPROM_LOCAL_ADDRESS)>1) {  
          if (valueD == HIGH && positionValue>1) {   
               moveDOWN();
               startMillis = millis();
               do { 
                    debouncer_down.update();
                    valueD=debouncer_down.read(); 
                    nowMillis = millis();
               } while ( valueD == HIGH 
                        && (nowMillis-startMillis)<OPERATION_TIME
                        && (positionValue-((nowMillis-startMillis)/OPERATION_TIME))>0 
                        && ((nowMillis-startMillis)/OPERATION_TIME)<100);
               saveState(EEPROM_LOCAL_ADDRESS, (uint8_t)(positionValue-(uint8_t)((nowMillis-startMillis)/OPERATION_TIME)));
               oldValueDW=valueD;
               moveSTOP();
             
          } else  {  
                DEBUG("Button DOWN released after ....", nowMillis-startMillis, "miliseconds" ); 
                somethingWrong();              
          } 
      }
  }  
}
void somethingWrong(void)
{
   positionValue = loadState(EEPROM_LOCAL_ADDRESS);  
   if (positionValue>100 || positionValue<0) { 
       //Something Wrong happened !! Closing blinds....
      moveDOWN();
      DEBUG("[Something Wrong happened] delay.....",OPERATION_TIME);
      delay(OPERATION_TIME);
      saveState(EEPROM_LOCAL_ADDRESS, 0);
      moveSTOP();
   }    
}
void moveUP(void)
{  
  digitalWrite(RELAY_ON_PIN, RELAY_ON);
  digitalWrite(RELAY_MV_PIN, RELAY_OFF);
}
void moveDOWN(void)
{
  digitalWrite(RELAY_ON_PIN, RELAY_ON);
  digitalWrite(RELAY_MV_PIN, RELAY_ON);  
}
void moveSTOP(void)
{
  digitalWrite(RELAY_MV_PIN, RELAY_OFF);   
  digitalWrite(RELAY_ON_PIN, RELAY_OFF);

  positionValue=loadState(EEPROM_LOCAL_ADDRESS);
  DEBUG("...... Sending from:", MY_NODE_ID," NEW blind position:",positionValue);
    // Send Blind Position to GW
  send(msgPos.set(positionValue)); // Send new state and NO request ack back
         
}

uint32_t lastSense;
inline bool updateGWinfo(void)
{
  long now = millis();
  if (now < lastSense || now - lastSense > SLEEP_TIME) {
    // READ DHT22
    sense();
    
    // READ LIGHT SENSOR, IF EXISTS
    if (existsSensor)  changeLight();
  
    // Send Blind Position to GW
    send(msgPos.set(loadState(EEPROM_LOCAL_ADDRESS))); // Send new state and NO request ack back
    
    // Write some debug info
    DEBUG("Sending from:", MY_NODE_ID," actual blind position:",loadState(EEPROM_LOCAL_ADDRESS));
    lastSense = now;
    return true;
  } else {
    return false;
  }
}       
void moveBlind(uint8_t newPosition)
{
  DEBUG("move blind .....", newPosition);
    positionValue = loadState(EEPROM_LOCAL_ADDRESS);
    long lActionTime = 0;
    if (newPosition==0) {
      moveDOWN();
      DEBUG("[Closing blind] delay.....",OPERATION_TIME);
      delay(OPERATION_TIME);
      saveState(EEPROM_LOCAL_ADDRESS, 0);
      moveSTOP();
    }
    else if (newPosition>5 && abs(newPosition-positionValue)>5) {
      lActionTime = (OPERATION_TIME/100)*(abs(positionValue - newPosition));       
      if (newPosition<positionValue) {
        DEBUG("delay.....",lActionTime);
        if (lActionTime>10) {
            moveDOWN();
            delay(lActionTime);
            saveState(EEPROM_LOCAL_ADDRESS,  (uint8_t)newPosition);
            moveSTOP();                       
        }
      } else if (newPosition>positionValue) {
        lActionTime = (OPERATION_TIME/100)*(newPosition - positionValue);  
        DEBUG("delay.....",lActionTime);
        if (lActionTime>10) {
            moveUP();
            delay(lActionTime);        
            saveState(EEPROM_LOCAL_ADDRESS,  (uint8_t)newPosition);
            moveSTOP(); 
        } 
      }            
    }    
}

void setup()  
{ 
  dht.setup(HUMIDITY_SENSOR_DIGITAL_PIN); 

  // Setup the buttons
  pinMode(BUTTON_UP_PIN,INPUT);
  pinMode(BUTTON_DOWN_PIN,INPUT);
  // Activate internal pull-up
  digitalWrite(BUTTON_UP_PIN,HIGH);
  digitalWrite(BUTTON_DOWN_PIN,HIGH);
  
  // After setting up the buttons, setup debouncers
  debouncer_up.attach(BUTTON_UP_PIN);
  debouncer_up.interval(5);
  debouncer_down.attach(BUTTON_DOWN_PIN);
  debouncer_down.interval(5);

  // Make sure relays are off when starting up
  digitalWrite(RELAY_ON_PIN, RELAY_OFF);
  digitalWrite(RELAY_MV_PIN, RELAY_OFF);
  // Then set relay pins in output mode
  pinMode(RELAY_ON_PIN, OUTPUT);   
  pinMode(RELAY_MV_PIN, OUTPUT);           

  DEBUG( "Node ", MY_NODE_ID," ready to receive messages..." );  
}
void presentation()  
{ 
  // Send the Sketch Version Information to the Gateway
  sendSketchInfo("test", "1.0");

  // Register all sensors to gw (they will be created as child devices)
  present(CHILD_ID_HUM, S_HUM);
  present(CHILD_ID_TEMP, S_TEMP);
  present(CHILD_ID_L, S_LIGHT);
  //present(CHILD_ID_REL_ON, S_LIGHT);
  present(CHILD_ID_REL_MV, S_DIMMER);
}

void loop()      
{   
  // listen UP button      
  upAction();
    
  // listen DOWN button      
  downAction();        
     
  //  Send data to GW
  updateGWinfo();

  // Check if blind position it's ok     
  somethingWrong();   
}

void receive(const MyMessage &message) {
  // We only expect one type of message from controller. But we better check anyway.
  if (message.isAck()) {
     DEBUG("This is an ack from gateway");
  } 

  if (message.type == V_LIGHT) {
    boolean state = message.getBool();
    DEBUG("V_LIGHT command received...",state); 
   
    if (state == 0) moveBlind(0);
    else if (state == 1) moveBlind(100);    
  }
  else if (message.type == V_DIMMER) {
    //int dimvalue= (atoi( message.data ))/10;
    uint8_t dimvalue= atoi( message.data );
    DEBUG( "V_DIMMER command received...", dimvalue );  
    if ((dimvalue<0)||(dimvalue>100)) {
      DEBUG( "V_DIMMER data invalid (should be 0..100)" );
      return;
    }
    if (dimvalue==0) {
      //moveBlind(0);
      DEBUG("received '0'.... don't do anything....");
    }
    else {
      moveBlind(dimvalue);
    }
  }
  else {
    DEBUG( "Invalid command received..." );  
    return;
  }
}


